/*	
 *	frammp3.c
 *
 *	P/ECE MP3 Driver
 *
 *	CLiP - Common Library for P/ECE
 *	Copyright (C) 2001-2003 Naoyuki Sawa
 *
 *	* Tue Oct 28 08:11:00 JST 2003 Naoyuki Sawa
 *	- 1st [XB
 */
#include "clip.h"
#include "clipmp3.h" /* TODO:Cuɑgݍ񂾂O܂ */

/****************************************************************************
 *	rbgXg[
 ****************************************************************************/

/* rbgXg[J܂B
 * [in]
 *	bs		rbgXg[\́B
 *	pos		ŏ̓ǂݍ݃AhXB
 */
void
mp3_bitstream_open(MP3BITSTREAM* bs, const void* pos)
{
	bs->cnt = 0;
	/*bs->buf = 0;{{sv}}*/
	bs->pos = (const unsigned char*)pos;
}

/* rbgXg[rbgǂݍ݂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	cnt		ǂݍރrbgB(0`32)
 * [out]
 *	߂l		ǂݍ񂾃rbgB
 */
int
mp3_bitstream_get(MP3BITSTREAM* bs, int cnt)
{
	int bits;

	ASSERT(0 <= cnt && cnt <= 32);

	if(cnt <= 24) {
		while(bs->cnt < cnt) {
			bs->buf = bs->buf << 8 | *bs->pos++;
			bs->cnt += 8;
		}
		bs->cnt -= cnt;
		bits = (bs->buf >> bs->cnt) & ((1 << cnt) - 1);
	} else {
		bits = mp3_bitstream_get(bs,       24) << 8 |
		       mp3_bitstream_get(bs, cnt - 24);
	}

	return bits;
}

/* rbgXg[rbgǂݔ΂܂B
 * [in]
 *	bs		rbgXg[\́B
 *	cnt		ǂݔ΂rbgB(0`C)
 */
void
mp3_bitstream_skip(MP3BITSTREAM* bs, int cnt)
{
	if(bs->cnt >= cnt) {
		bs->cnt -= cnt;
	} else {
		cnt -= bs->cnt;
		bs->pos += cnt / 8;
		cnt &= 7;
		if(cnt) {
			bs->buf = *bs->pos++;
			bs->cnt = 8 - cnt;
		} else {
			bs->cnt = 0;
		}
	}
}

/****************************************************************************
 *	nt}
 ****************************************************************************/

/* BigValuë̒PRegion𕜍Tu[`łB
 * [in]
 *	bs		rbgXg[\́B
 *	i_table		e[uԍB
 *	count		TvB
 *	out		o̓obt@B
 * [note]
 *	* ߖ̂߂ɁATu[`ɕ܂B
 *	  Amp3_huffman_decode()̒3R[hWJƁA300oCgĂ܂܂B
 *	* ̊֐1Oj[3Ăяo邾Ȃ̂ŁAقǑxቺ͂ȂƎv܂B
 */
static void
mp3_huffman_decode_region(MP3BITSTREAM* bs, int i_table, int count, short* out)
{
	int linbits, i_node, x, y;
	const short (*table)[2];

	/* e[u擾B */
	table   = mp3_huffman_table[i_table].table;
	linbits = mp3_huffman_table[i_table].linbits;

	/* B */
	if(table) {
		while(count) {
			i_node = 0;
			do {
				i_node = table[i_node][mp3_bitstream_get(bs, 1)];
			} while(i_node > 0);
			i_node = -i_node;
			x = i_node >> 4;
			y = i_node & 15;
			if(x == 15) x += mp3_bitstream_get(bs, linbits);
			if(x && mp3_bitstream_get(bs, 1)) x = -x;
			if(y == 15) y += mp3_bitstream_get(bs, linbits);
			if(y && mp3_bitstream_get(bs, 1)) y = -y;
			*out++ = x;
			*out++ = y;
			count -= 2/*{x,y}*/;
		}
	} else {
		/* [Œe[uB(SCMPX͐܂񂪁ALAME͐܂) */
		while(count) {
			*out++ = 0;
			*out++ = 0;
			count -= 2/*{x,y}*/;
		}
	}
}

/* nt}B
 * [in]
 *	bs		rbgXg[\́B
 *	granule		Oj[\́B
 *	out		o̓obt@B
 */
void
mp3_huffman_decode(MP3BITSTREAM* bs, const MP3FRAMEHEADER* frame, const MP3GRANULEINFO* granule, short* out/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/)
{
	int i, count, region1_start, region2_start, big_values_2;

	/* Region0/1/2̋E擾B */
	if(granule->block_type == 2) {
		/* V[gubN(~bNXubN?) */
		region1_start =  36;
		region2_start = 576;
	} else {
		/* OubN */
		region1_start = mp3_scalefactor_band_index_table[frame->sampling_frequency][0/*LongBlock*/][(granule->region0_count + 1)];
		region2_start = mp3_scalefactor_band_index_table[frame->sampling_frequency][0/*LongBlock*/][(granule->region0_count + 1) + (granule->region1_count + 1)];
	}

	/* BigValuë... */
	big_values_2 = granule->big_values * 2/*{x,y}*/;
	if(region1_start > big_values_2) region1_start = big_values_2; /* Kv */
	if(region2_start > big_values_2) region2_start = big_values_2; /* Kv */
	i = 0;
	/* Region0 */
	count = region1_start - i;
	mp3_huffman_decode_region(bs, granule->table_select[0], count, out);
	i   += count;
	out += count;
	/* Region1 */
	count = region2_start - i;
	mp3_huffman_decode_region(bs, granule->table_select[1], count, out);
	i   += count;
	out += count;
	/* Region2 */
	count = big_values_2  - i;
	mp3_huffman_decode_region(bs, granule->table_select[2], count, out);
	i   += count;
	out += count;

	/* Count1̈... */
	/* ȗ */

	/* rzerö... */
	while(i < MP3_GRANULE_SIZE) {
		*out++ = 0;
		i++;
	}
}

/****************************************************************************
 *	tʎq
 ****************************************************************************/

/* * MP3dlɒ߂ꂽAtʎq̌vZ͎̒ʂłB
 *
 *	xr  is ~ |is|^(1/3) ~ 2^(Gain-Scale)
 *
 *	is: tʎqւ̓̓f[^
 *	xr: tʎq̏o͌
 *
 *   A
 *
 *	Gain   (GlobalGain-210)/4 - SubblockGain
 *	Scale  ((ScalefactorScale+1)~(Scalefactor+Preemphasis)) / 2
 *
 *   SubblockGainAScalefactorScaleAPreemphais̏ȗ0Ɖ肷ƁA
 *
 *	Gain   (GlobalGain-210) / 4
 *	Scale  Scalefactor / 2
 *
 *   ܂Ƃ߂ƁAZł͎̂悤ɏł܂B
 *
 *	xr  is ~ |is|^(1/3) ~ 2^((GlobalGain-210)/4 - Scalefactor/2)
 *	    is ~ |is|^(1/3) ~ 2^((GlobalGain-210)/4)  2^(Scalefactor/2)
 *	    is ~ |is|^(1/3) ~ 2^((GlobalGain-210)/4) >> (Scalefactor/2)
 *	            ~~~~~~~~~~    ~~~~~~~~~~~~~~~~~~~~~~
 *	            e[u          e[u      
 */

/* tʎqB(OubN)
 * [in]
 *	frame		t[wb_\́B
 *	granule		Oj[\́B
 *	scale		XP[t@N^\́B
 *	in		̓obt@B
 *	out		o̓obt@B
 * [note]
 *	* XP[t@N^XP[(MP3GRANULEINFO.scalefac_scale)͖ΉłB
 *	  ۂMP3t@Cł͂قƂ0ɂȂĂ悤Ȃ̂ŁA܂e͂ȂƎv܂B
 */
void
mp3_requantize_long(const MP3FRAMEHEADER* frame, const MP3GRANULEINFO* granule, const MP3SCALEFACTOR* scale, const short* in/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/, short* out/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/)
{
	int gain, i, w, index;
	short h;
	const short* sfb_index;
	const short* sfb_scale;

	/* XP[t@N^oh̏B */
	sfb_index = &mp3_scalefactor_band_index_table[frame->sampling_frequency][0/*LongBlock*/][1/*̃oh̊JnCfNX*/];
	sfb_scale = scale->long_block;
	index = 0;

	gain = mp3_gain_table[granule->global_gain];	/* 1.13.18 */
	for(i = 0; i < MP3_GRANULE_SIZE; i++) {
		/* g1TvtʎqB */
		h = *in++;				/* 1.15. 0 */
		w = h * mp3_pow1_3_table[abs(h)];	/* 1.21.10 (short~short) */
		w = w * gain;				/* 1. 3.28 (int~int) */
		*out++ = w >> (14 + *sfb_scale / 2);	/* 1. 1.14 (I[o[t[) */

		/* ̃XP[t@N^oh̊JnCfNXɓBAgpXP[֐i߂܂B */
		index++;
		if(index == *sfb_index) {
			sfb_index++;
			sfb_scale++;
		}
	}
}

/* tʎqB(V[gubN)
 * [in]
 *	frame		t[wb_\́B
 *	granule		Oj[\́B
 *	scale		XP[t@N^\́B
 *	in		̓obt@B
 *	out		o̓obt@B
 * [note]
 *	* XP[t@N^XP[(MP3GRANULEINFO.scalefac_scale)͖ΉłB
 *	  ۂMP3t@Cł͂قƂ0ɂȂĂ悤Ȃ̂ŁA܂e͂ȂƎv܂B
 *	* TuubNQC(MP3GRANULEINFO.subblock_gain[])͖ΉłB
 *	  ۂMP3t@Cł͂قƂ0ɂȂĂ悤Ȃ̂ŁA܂e͂ȂƎv܂B
 *	* ʏMP3t@Cł̓OubN唼߁AV[gubN͑Ŝ10%xłB
 *	  łAV[gubN֐ĂAb͏ȂłB
 *	  RAMȂȂꍇ́Å֐SRAMɈڂĂ܂ĂƎv܂B
 */
void
mp3_requantize_short(const MP3FRAMEHEADER* frame, const MP3GRANULEINFO* granule, const MP3SCALEFACTOR* scale, const short* in/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/, short* out/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/)
{
	int gain, i, j, k, w, index;
	short h;
	const short* sfb_index;
	const short (*sfb_scale)[MP3_SUBBLOCK_COUNT];
	const short* reorder;

	/* XP[t@N^oh̏B */
	sfb_index = &mp3_scalefactor_band_index_table[frame->sampling_frequency][1/*ShortBlock*/][1/*̃oh̊JnCfNX*/];
	sfb_scale = scale->short_block;
	index = 0;

	/* TvёւṕAǂݍݑCfNXe[u擾B */
	reorder = mp3_reorder_read_index_table[frame->sampling_frequency];

	gain = mp3_gain_table[granule->global_gain];	/* 1.13.18 */
	for(i = 0; i < MP3_SUBBAND_COUNT; i++) {
		for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
			/* g1AԎ3Tv(=3TuubN)tʎqB
			 * ̌AV[gubNIMDCT (g6~Ԏ1)~Ԏ3 PʂƂď̂ŁA
			 *  [g32][Ԏ3][g6] ̏ɕёւĂ܂B
			 */
			for(k = 0; k < MP3_SUBBLOCK_COUNT; k++) {
				h = in[*reorder++];			/* 1.15. 0 */
				w = h * mp3_pow1_3_table[abs(h)];	/* 1.21.10 (short~short) */
				w = w * gain;				/* 1. 3.28 (int~int) */
				*out = w >> (14 + (*sfb_scale)[k] / 2);	/* 1. 1.14 (I[o[t[) */
				out += MP3_SHORT_IMDCT_SIZE;
			}
			out -= MP3_SUBBAND_SIZE - 1;

			/* ̃XP[t@N^oh̊JnCfNXɓBAgpXP[֐i߂܂B */
			index++;
			if(index == *sfb_index) {
				sfb_index++;
				sfb_scale++;
			}
		}
		out += MP3_SUBBAND_SIZE - MP3_SHORT_IMDCT_SIZE;
	}
}

/****************************************************************************
 *	IMDCT
 ****************************************************************************/

/* IMDCTB(OubN)
 * [in]
 *	in		̓obt@B
 *	out		o̓obt@B
 *	save		t[ԂŕۑKv̂obt@B
 * [note]
 *	* Tuoḧꕔ肵Ă܂B
 */
void
mp3_imdct_long(const short* in/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/, short* out/*[MP3_SUBBAND_SIZE][MP3_SUBBAND_COUNT]*/, short* save/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/)
{
#ifndef PIECE /**************************************************************/

	int i, j, k, sum;
	const short* m;

	for(k = 0; k < MP3_SUBBAND_COUNT; k++) {
		m = &mp3_imdct_long_matrix[0][0];

		/* uÕOj[̌㔼vƁũ݂Oj[̑OvďóB */
		for(i = 0; i < MP3_SUBBAND_SIZE; i++) {
			sum = 0;
			for(j = 0; j < MP3_SUBBAND_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			sum += *save++;	/* 1.1.14 */
			/* TuoḧꕔBs𕄍]ē]uĂB */
			if(k & i & 1) sum = -sum;
			*out = sum;	/* 1.1.14 */
			out += MP3_SUBBAND_COUNT;
			in  -= MP3_SUBBAND_SIZE;
		}

		out  -= MP3_GRANULE_SIZE - 1;
		save -= MP3_SUBBAND_SIZE;

		/* ũOj[̑Ovɍ邽߂Ɂũ݂Oj[̌㔼vۑB */
		for(i = 0; i < MP3_SUBBAND_SIZE; i++) {
			sum = 0;
			for(j = 0; j < MP3_SUBBAND_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			*save++ = sum;	/* 1.1.14 */
			in -= MP3_SUBBAND_SIZE;
		}

		in += MP3_SUBBAND_SIZE;
	}

#else	/********************************************************************/

	asm volatile ("
	xld.w %r7, 32				; k = MP3_SUBBAND_COUNT
mp3_imdct_long_L10:
	xld.w %r11, mp3_imdct_long_matrix	; m = mp3_imdct_long_matrix
	;
	; uÕOj[̌㔼vƁũ݂Oj[̑OvďóB
	;
	xld.w %r6, 18				; i = MP3_SUBBAND_SIZE
mp3_imdct_long_L20:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 18				; MP3_SUBBAND_SIZE
	mac %r10	
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h %r15, [%r14]+			; sum += *save++
	add %r10, %r15
	;
	xld.w %r15, 1				; if(k & i & 1) sum = -sum
	and %r15, %r7
	and %r15, %r6
	xjreq mp3_imdct_long_L30
	not %r10, %r10
	xadd %r10, %r10, 1
mp3_imdct_long_L30:
	ld.h [%r13], %r10			; *out = sum
	xadd %r13, %r13, 64			; out += MP3_SUBBAND_COUNT * sizeof(short)
	xsub %r12, %r12, 36			; in -= MP3_SUBBAND_SIZE * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_long_L20
	;
	xsub %r13, %r13, 1150			; out -= (MP3_GRANULE_SIZE - 1) * sizeof(short)
	xsub %r14, %r14, 36			; save -= MP3_SUBBAND_SIZE * sizeof(short)
	;
	; ũOj[̑Ovɍ邽߂Ɂũ݂Oj[̌㔼vۑB
	;
	xld.w %r6, 18				; i = MP3_SUBBAND_SIZE
mp3_imdct_long_L40:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 18				; MP3_SUBBAND_SIZE
	mac %r10	
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h [%r14]+, %r10			; *save++ = sum
	xsub %r12, %r12, 36			; in -= MP3_SUBBAND_SIZE * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_long_L40
	;
	xadd %r12, %r12, 36			; in += MP3_SUBBAND_SIZE * sizeof(short)
	;
	xsub %r7, %r7, 1
	xjrne mp3_imdct_long_L10
	");

#endif	/********************************************************************/
}

void
mp3_imdct_short(const short* in/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/, short* out/*[MP3_SUBBAND_SIZE][MP3_SUBBAND_COUNT]*/, short* save/*[MP3_SUBBAND_COUNT][MP3_SUBBAND_SIZE]*/)
{
#ifndef PIECE /**************************************************************/

	int i, j, k, sum;
	const short* m;

	for(k = 0; k < MP3_SUBBAND_COUNT; k++) {
		/* |=`=| |=a=| |=b=| |=c=| |=d=| |=e=| */
		/* 000000 ------ ------ ------ ------ 000000 */
		/* ------ ?????? ?????? ------ ------ ------ */
		/* ------ ------ ?????? ?????? ------ ------ */
		/* ------ ------ ------ ?????? ?????? ------ */
		/* |=saveƑďo=| |====saveɕۑ====| ~ MP3_SUBBAND_COUNT */

		/* |=`=| */
		for(i = 0; i < MP3_SHORT_IMDCT_SIZE; i++) {
			sum = *save++;	/* 1.1.14 */
			/* TuoḧꕔBs𕄍]ē]uĂB */
			if(k & i & 1) sum = -sum;
			*out = sum;	/* 1.1.14 */
			out += MP3_SUBBAND_COUNT;
		}

		/* |=a=| */
		m = &mp3_imdct_short_matrix[0][0];
		for(i = 0; i < MP3_SHORT_IMDCT_SIZE; i++) {
			sum = 0;
			for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			sum += *save++;	/* 1.1.14 */
			/* TuoḧꕔBs𕄍]ē]uĂB */
			if(k & i & 1) sum = -sum;
			*out = sum;	/* 1.1.14 */
			out += MP3_SUBBAND_COUNT;
			in  -= MP3_SHORT_IMDCT_SIZE;
		}

		/* |=b=| */
		for(i = 0; i < MP3_SHORT_IMDCT_SIZE; i++) {
			sum = 0;
			for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			m -= MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE;
			for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			sum += *save++;	/* 1.1.14 */
			/* TuoḧꕔBs𕄍]ē]uĂB */
			if(k & i & 1) sum = -sum;
			*out = sum;	/* 1.1.14 */
			out += MP3_SUBBAND_COUNT;
			m  += MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE;
			in -= MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE;
		}
		m    -= MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE;
		in   += MP3_SHORT_IMDCT_SIZE;
		out  -= MP3_GRANULE_SIZE - 1;
		save -= MP3_SUBBAND_SIZE;

		/* |=c=| */
		for(i = 0; i < MP3_SHORT_IMDCT_SIZE; i++) {
			sum = 0;
			for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			m -= MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE;
			for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			*save++ = sum;	/* 1.1.14 */
			m  += MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE;
			in -= MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE;
		}
		m  -= MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE;
		in += MP3_SHORT_IMDCT_SIZE;

		/* |=d=| */
		for(i = 0; i < MP3_SHORT_IMDCT_SIZE; i++) {
			sum = 0;
			for(j = 0; j < MP3_SHORT_IMDCT_SIZE; j++) {
				sum += *in++ * *m++;	/* short~short 1.3.28 */
			}
			sum >>= 14;	/* 1.1.14 */
			*save++ = sum;	/* 1.1.14 */
			in -= MP3_SHORT_IMDCT_SIZE;
		}
		in += MP3_SHORT_IMDCT_SIZE;

		/* |=e=| */
		for(i = 0; i < MP3_SHORT_IMDCT_SIZE; i++) {
			*save++ = 0;	/* 1.1.14 */
		}
	}

#else	/********************************************************************/

	/*  */
	asm volatile ("
	xld.w %r7, 32				; k = MP3_SUBBAND_COUNT
mp3_imdct_short_L10:
	;
	; |=`=|
	;
	xld.w %r6, 6				; i = MP3_SHORT_IMDCT_SIZE
mp3_imdct_short_L20:
	ld.h %r10, [%r14]+			; sum = *save++
	;
	xld.w %r15, 1				; if(k & i & 1) sum = -sum
	and %r15, %r7
	and %r15, %r6
	xjreq mp3_imdct_short_L30
	not %r10, %r10
	xadd %r10, %r10, 1
mp3_imdct_short_L30:
	ld.h [%r13], %r10			; *out = sum
	xadd %r13, %r13, 64			; out += MP3_SUBBAND_COUNT * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_short_L20
	;
	; |=a=|
	;
	xld.w %r11, mp3_imdct_short_matrix	; m = mp3_imdct_short_matrix
	xld.w %r6, 6				; i = MP3_SHORT_IMDCT_SIZE
mp3_imdct_short_L40:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 6				; MP3_SHORT_IMDCT_SIZE
	mac %r10	
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h %r15, [%r14]+			; sum += *save++
	add %r10, %r15
	;
	xld.w %r15, 1				; if(k & i & 1) sum = -sum
	and %r15, %r7
	and %r15, %r6
	xjreq mp3_imdct_short_L50
	not %r10, %r10
	xadd %r10, %r10, 1
mp3_imdct_short_L50:
	ld.h [%r13], %r10			; *out = sum
	xadd %r13, %r13, 64			; out += MP3_SUBBAND_COUNT * sizeof(short)
	xsub %r12, %r12, 12			; in -= MP3_SHORT_IMDCT_SIZE * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_short_L40
	;
	; |=b=|
	;
	xld.w %r6, 6				; i = MP3_SHORT_IMDCT_SIZE
mp3_imdct_short_L60:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 6				; MP3_SHORT_IMDCT_SIZE
	mac %r10	
	xsub %r11, %r11, 84			; m -= (MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	; (%alrp)				; ݍ
	xld.w %r10, 6				; MP3_SHORT_IMDCT_SIZE
	mac %r10	
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h %r15, [%r14]+			; sum += *save++
	add %r10, %r15
	;
	xld.w %r15, 1				; if(k & i & 1) sum = -sum
	and %r15, %r7
	and %r15, %r6
	xjreq mp3_imdct_short_L70
	not %r10, %r10
	xadd %r10, %r10, 1
mp3_imdct_short_L70:
	ld.h [%r13], %r10			; *out = sum
	xadd %r13, %r13, 64			; out += MP3_SUBBAND_COUNT * sizeof(short)
	xadd %r11, %r11, 72			; m += (MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	xsub %r12, %r12, 24			; in -= (MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_short_L60
	;
	xsub %r11, %r11, 72			; m -= (MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	xadd %r12, %r12, 12			; in += MP3_SHORT_IMDCT_SIZE * sizeof(short)
	xsub %r13, %r13, 1150			; out -= (MP3_GRANULE_SIZE - 1) * sizeof(short)
	xsub %r14, %r14, 36			; save -= MP3_SUBBAND_SIZE * sizeof(short)
	;
	; |=c=|
	;
	xld.w %r6, 6				; i = MP3_SHORT_IMDCT_SIZE
mp3_imdct_short_L80:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 6				; MP3_SHORT_IMDCT_SIZE
	mac %r10	
	xsub %r11, %r11, 84			; m -= (MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	; (%alrp)				; ݍ
	xld.w %r10, 6				; MP3_SHORT_IMDCT_SIZE
	mac %r10	
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h [%r14]+, %r10			; *save++ = sum
	xadd %r11, %r11, 72			; m += (MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	xsub %r12, %r12, 24			; in -= (MP3_SHORT_IMDCT_SIZE + MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_short_L80
	;
	xsub %r11, %r11, 72			; m -= (MP3_SHORT_IMDCT_SIZE * MP3_SHORT_IMDCT_SIZE) * sizeof(short)
	xadd %r12, %r12, 12			; in += MP3_SHORT_IMDCT_SIZE * sizeof(short)
	;
	; |=d=|
	;
	xld.w %r6, 6				; i = MP3_SHORT_IMDCT_SIZE
mp3_imdct_short_L90:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 6				; MP3_SHORT_IMDCT_SIZE
	mac %r10	
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h [%r14]+, %r10			; *save++ = sum
	xsub %r12, %r12, 12			; in -= MP3_SHORT_IMDCT_SIZE * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_short_L90
	;
	xadd %r12, %r12, 12			; in += MP3_SHORT_IMDCT_SIZE * sizeof(short)
	;
	; |=e=|
	;
	xld.w %r6, 6				; i = MP3_SHORT_IMDCT_SIZE
mp3_imdct_short_L100:
	ld.h [%r14]+, %r8			; *save++ = 0
	;
	xsub %r6, %r6, 1
	xjrne mp3_imdct_short_L100
	;
	xsub %r7, %r7, 1
	xjrne mp3_imdct_short_L10
	");

#endif	/********************************************************************/
}

/****************************************************************************
 *	Tuoh
 ****************************************************************************/

/* TuohB
 * [in]
 *	in		̓obt@B
 *	out		o̓obt@B
 *	save		t[ԂŕۑKv̂obt@B
 *	offset		saveItZbgB
 *			Omp3_subband_synthesys()̖߂lnĂB
 * [out]
 *	߂l		saveItZbgB
 * [note]
 *	* Tuoḧꕔmp3_imdct()ŐɍsĂ܂B
 */
int
mp3_subband_synthesys(const short* in/*[MP3_SUBBAND_SIZE][MP3_SUBBAND_COUNT]*/, short* out/*[MP3_SUBBAND_SIZE][MP3_SUBBAND_COUNT_HALF]*/, short* save/*[16][MP3_SUBBAND_COUNT_2_HALF]*/, int offset)
{
#ifndef PIECE /**************************************************************/

	int i, j, k, sum;
	const short* m;

#define OFFSET_MASK	(MP3_SUBBAND_COUNT_2_HALF * 16 - 1)

	for(k = 0; k < MP3_SUBBAND_SIZE; k++) {
		/* ݍ݉ZB(32f[^ԗ̈ɕϊ)
		 * ͍s̊s𕄍]ē]u鏈́AIMDCT̒ōsĂ܂B
		 */
		m = &mp3_subband_matrix[0][0];
		for(i = 0; i < MP3_SUBBAND_COUNT_2_HALF; i++) {
			sum = 0;
			for(j = 0; j < MP3_SUBBAND_COUNT; j++) {
				sum += *in++ * *m++;	/* 1.3.28 */
			}
			save[offset++] = sum >> 14;	/* 1.1.14 */
			in -= MP3_SUBBAND_COUNT;
		}
		offset -= MP3_SUBBAND_COUNT_2_HALF;
		in += MP3_SUBBAND_COUNT;

		/* 16g̃f[^̏dˍ킹B
		 * |tF[YtB^[oNW\́A炩ߓ]uĂ܂B
		 */
		m = &mp3_polyphase_filter[0][0];
		for(i = 0; i < MP3_SUBBAND_COUNT_HALF; i++) {
			sum = 0;
			for(j = 0; j < 16 / 2/*2TvÂ*/; j++) {
				sum += save[offset] * *m++;	/* 1.3.28 */
				offset += MP3_SUBBAND_COUNT_2_HALF + MP3_SUBBAND_COUNT_HALF;
				offset &= OFFSET_MASK;
				sum += save[offset] * *m++;	/* 1.3.28 */
				offset += MP3_SUBBAND_COUNT_2_HALF - MP3_SUBBAND_COUNT_HALF;
				offset &= OFFSET_MASK;
			}
			offset++;
			/* -0x7fff`0x7fffɃXP[OB */
			sum >>= 13;	/* 1.(16).15 */
			if(sum < -0x7fff) sum = -0x7fff;
			if(sum >  0x7fff) sum =  0x7fff;
			*out++ = sum;	/* 1.15 */
		}
		offset -= MP3_SUBBAND_COUNT_2_HALF + MP3_SUBBAND_COUNT_HALF;
		offset &= OFFSET_MASK;
	}

#undef OFFSET_MASK

#else	/********************************************************************/

	asm volatile ("
	xld.w %r7, 18				; k = MP3_SUBBAND_SIZE
mp3_subband_synthesys_L10:
	;
	; ݍ݉ZB(32f[^ԗ̈ɕϊ)
	; ͍s̊s𕄍]ē]u鏈́AIMDCT̒ōsĂ܂B
	;
	xld.w %r11, mp3_subband_matrix		; m = mp3_subband_matrix
	xld.w %r6, 32				; i = MP3_SUBBAND_COUNT_2_HALF
	ld.w %r5, %r14				; save_ptr = &save[offset]
	add %r5, %r15
mp3_subband_synthesys_L20:
	ld.w %alr, %r8				; ݍ
	xld.w %r10, 32				; MP3_SUBBAND_COUNT
	mac %r10
	ld.w %r10, %alr				; sum >>= 14
	xsra %r10, 14
	ld.h [%r5]+, %r10			; *save_ptr++ = sum
	xsub %r12, %r12, 64			; in -= MP3_SUBBAND_COUNT * sizeof(short)
	;
	xsub %r6, %r6, 1
	xjrne mp3_subband_synthesys_L20
	;
	xadd %r12, %r12, 64			; in += MP3_SUBBAND_COUNT * sizeof(short)
	;
	; 16g̃f[^̏dˍ킹B
	; |tF[YtB^[oNW\́A炩ߓ]uĂ܂B
	;
	xld.w %r6, mp3_polyphase_filter		; m = mp3_polyphse_filter
	xld.w %r11, 16				; i = MP3_SUBBAND_COUNT_HALF
mp3_subband_synthesys_L30:
	ld.w %alr, %r8				; sum = 0
	xld.w %r10, 8				; j = 16 / 2/*2TvÂ*/
mp3_subband_synthesys_L40:
	ld.w %r5, %r14				; sum += save[offset] * *m++
	add %r5, %r15
	xld.w %r4, 1
	mac %r4
	xadd %r15, %r15, 96			; offset += (MP3_SUBBAND_COUNT_2_HALF + MP3_SUBBAND_COUNT_HALF) * sizeof(short)
	xand %r15, %r15, 1023			; offset &= OFFSET_MASK (= MP3_SUBBAND_COUNT_2_HALF * 16 * sizeof(short) - 1)
	ld.w %r5, %r14				; sum += save[offset] * *m++
	add %r5, %r15
	xld.w %r4, 1
	mac %r4
	xadd %r15, %r15, 32			; offset += (MP3_SUBBAND_COUNT_2_HALF - MP3_SUBBAND_COUNT_HALF) * sizeof(short)
	xand %r15, %r15, 1023			; offset &= OFFSET_MASK (= MP3_SUBBAND_COUNT_2_HALF * 16 * sizeof(short) - 1)
	;
	xsub %r10, %r10, 1
	xjrne mp3_subband_synthesys_L40
	;
	xadd %r15, %r15, 2			; offset += 1 * sizeof(short)
	ld.w %r4, %alr				; sum >>= 13
	xsra %r4, 13
	xcmp %r4, -32767			; if(sum < -0x7fff) sum = -0x7fff
	xjrge mp3_subband_synthesys_L50
	xld.w %r4, -32767
mp3_subband_synthesys_L50:
	xcmp %r4, 32767				; if(sum >  0x7fff) sum =  0x7fff
	xjrle mp3_subband_synthesys_L60
	xld.w %r4, 32767
mp3_subband_synthesys_L60:
	ld.h [%r13]+, %r4			; *out++ = sum
	;
	xsub %r11, %r11, 1
	xjrne mp3_subband_synthesys_L30
	;
	xsub %r15, %r15, 96			; offset -= (MP3_SUBBAND_COUNT_2_HALF + MP3_SUBBAND_COUNT_HALF) * sizeof(short)
	xand %r15, %r15, 1023			; offset &= OFFSET_MASK (= MP3_SUBBAND_COUNT_2_HALF * 16 * sizeof(short) - 1)
	;
	xsub %r7, %r7, 1
	xjrne mp3_subband_synthesys_L10
	;
	ld.w %r10, %r15
	ret
	");

#endif	/********************************************************************/

	return offset;
}

